class LevelType:
    FLAT = "flat"
    DEFAULT = "default"

class GameMode:
    SURVIVAL = "survival"
    CREATIVE = "creative"
    SPECTATOR = "spectator"
    ADVENTURE = "adventure"

class Difficulty:
    PEACEFUL = "peaceful"
    EASY = "easy"
    NORMAL = "normal"
    HARD = "hard"

class Dimension:
    NETHER = "the_nether"
    OVERWORLD = "overworld"
    END = "the_end"

class BrandChannel:
    PREFIXED = "MC|Brand"
    IDENTIFIER = "minecraft:brand"

class ChatSetting:
    ENABLED = "enabled"
    COMMANDS_ONLY = "commandsOnly"
    DISABLED = "disabled"

class ViewDistance:
    FAR = "far"
    NORMAL = "normal"
    SHORT = "short"
    TINY = "tiny"